@@ -27,12 +27,23 @@ from pyts2.time import *
 from pyts2.utils import *
 from pyts2.filelock import FileLock
 
+def get_instant_or_fn(path):
+    try:
+        return str(TSInstant.from_path(path))
+    except ValueError:
+        return path
+
 
 def path_is_timestream_file(path, extensions=None):
-    """Test if pathname pattern matches the expected
+    """Test if path given contains a valid timestream format, and optionally
+    checks that it has an expected file extension.
 
     :param path: File path, with or without directory
-    :param path: Optionally, one or more extensions to accept
+    :type path: str
+    :param extensions: Optionally, one or more extensions to accept
+    :type extensions: str
+    :return: True if path is timestream and extension compatible, otherwise False
+    :rtype: bool
 
     >>> path_is_timestream_file("test_2018_12_31_23_59_59_00.jpg")
     True
@@ -72,8 +83,17 @@ def path_is_timestream_file(path, extensions=None):
 
 
 class Fetcher(object):
+    """ Gets files from archive bundles.
+    """
     @classmethod
     def from_json(self, obj):
+        """ Takes an object path and returns format-decoded file content.
+
+        :param obj: File object to look up
+        :type obj: JSON object/dict
+        :return: Fetched file
+        :rtype: File content
+        """
         if obj["type"] == "zip":
             return ZipContentFetcher(obj["archivepath"], obj["pathinzip"])
         elif obj["type"] == "tar":
@@ -81,17 +101,34 @@ class Fetcher(object):
 
     @property
     def instant(self):
+        """ Gets datetime from file path.
+
+        :return: Datetime
+        :rtype: :class:`.TSInstant`
+        """
         return TSInstant.from_path(self.filename)
 
 
 class ZipContentFetcher(Fetcher):
+    """ Retrieves files from zip archives.
+    """
     _fetchtype = 'zip'
 
     def __init__(self, archivepath, pathinzip):
+        """
+        :param archivepath: Path to zip file
+        :type archivepath: str
+        :param pathinzip: Path to file to retrieve, within zip
+        :type pathinzip: str
+        """
         self.archivepath = archivepath
         self.pathinzip = pathinzip
 
     def get(self):
+        """
+        :return: File retrieved from zip archive.
+        :rtype: File content
+        """
         with zipfile.ZipFile(str(self.archivepath)) as zfh:
             return zfh.read(self.pathinzip)
 
@@ -100,19 +137,35 @@ class ZipContentFetcher(Fetcher):
         return op.basename(self.pathinzip)
 
     def dict(self):
+        """
+        :return: Summary of fetch parameters
+        :rtype: dict
+        """
         return {"type": "zip",
                 "archivepath": self.archivepath,
                 "pathinzip": self.pathinzip}
 
 
 class TarContentFetcher(Fetcher):
+    """ Retrieves files from tar archives.
+    """
     _fetchtype = 'tar'
 
     def __init__(self, archivepath, pathintar):
+        """
+        :param archivepath: Path to tar file
+        :type archivepath: str
+        :param pathinzip: Path to file to retrieve, within tar file
+        :type pathinzip: str
+        """
         self.archivepath = archivepath
         self.pathintar = pathintar
 
     def get(self):
+        """
+        :return: File retrieved from tar archive.
+        :rtype: File content
+        """
         with tarfile.TarFile(self.archivepath) as tfh:
             return tfh.extractfile(self.pathintar).read()
 
@@ -121,16 +174,31 @@ class TarContentFetcher(Fetcher):
         return op.basename(self.pathintar)
 
     def dict(self):
+        """
+        :return: Summary of fetch parameters
+        :rtype: dict
+        """
         return {"type": "tar",
                 "archivepath": self.archivepath,
                 "pathintar": self.pathintar}
 
 
 class FileContentFetcher(Fetcher):
+    """ Retrieves files from disk, not in an archive.
+    """
+
     def __init__(self, path):
+        """
+        :param path: Path to file to retrieve
+        :type path: str
+        """
         self.pathondisk = Path(path)
 
     def get(self):
+        """
+        :return: File retrieved from disk.
+        :rtype: File content
+        """
         with open(self.pathondisk, "rb") as fh:
             return fh.read()
 
@@ -139,6 +207,10 @@ class FileContentFetcher(Fetcher):
         return op.basename(self.pathondisk)
 
     def dict(self):
+        """
+        :return: Summary of fetch parameters
+        :rtype: dict
+        """
         return {"type": "file",
                 "path": self.pathondisk}
 
@@ -147,6 +219,20 @@ class TimestreamFile(object):
     '''A container class for files in timestreams'''
 
     def __init__(self, instant=None, filename=None, fetcher=None, content=None, report=None, format=None):
+        """
+        :param instant: Datetime point that this file represents
+        :param filename: Name of file, can be retrieved from fetcher
+        :param fetcher: File fetcher object to retrieve file content from disk or within archive formats
+        :param content: File content (usually if creating timestream file in pipelines)
+        :param report: Variable to store reports from Timestream pipeline components
+        :param format: File format, often the file extension (e.g. ``jpg``)
+        :type instant: TSInstant object or derived from file path
+        :type filename: str
+        :type fetcher: :class:`.Fetcher`, optional
+        :type content: File content, optional
+        :type report: dict
+        :type format: str, optional
+        """
         self.instant = instant
         self.filename = filename
         self.fetcher = fetcher
@@ -167,7 +253,7 @@ class TimestreamFile(object):
             raise ValueError("TimestreamFile must have content (directly or via a fetcher)")
         if format is None:
             format = op.splitext(self.filename)[1]
-        self.format = re.sub(r'^\.+', '', format)
+        self.format = format.lower().lstrip(".")
 
     @property
     def content(self):
@@ -178,6 +264,8 @@ class TimestreamFile(object):
         return self._content
 
     def clear_content(self):
+        """ Deletes all file content in object.
+        """
         del self._content
         self._content = None
         if hasattr(self, '_pixels'):
@@ -192,6 +280,15 @@ class TimestreamFile(object):
 
     @classmethod
     def from_path(cls, path, instant=None):
+        """ Create timestream file from file path.
+
+        :param path: Path to file
+        :type path: str
+        :param instant: Timepoint associated with object. Attempts to use datetime in path if not provided.
+        :type instant: :class:`time.TSInstant`, optional
+        :return: Timestream file
+        :rtype: :class:`.TimestreamFile`
+        """
         if instant is None:
             instant = TSInstant.from_path(path)
         return cls(fetcher=FileContentFetcher(path),
@@ -200,6 +297,17 @@ class TimestreamFile(object):
 
     @classmethod
     def from_bytes(cls, filebytes, filename, instant=None):
+        """ Create timestream file programmatically.
+
+        :param filebytes: Bytes to encode
+        :type filebytes: bytes
+        :param filename: Name of file
+        :type filename: str
+        :param instant: Timepoint associated with object. Attempts to use datetime in path if not provided.
+        :type instant: :class:`time.TSInstant`, optional
+        :return: Timestream file
+        :rtype: :class:`.TimestreamFile`
+        """
         if not isinstance(filebytes, bytes):
             raise ValueError("from_bytes must be given file contents as bytes")
         if instant is None:
@@ -214,6 +322,13 @@ class TimestreamFile(object):
         return len(self.content)
 
     def checksum(self, algorithm="md5"):
+        """ Checksum content of this object by the given algorithm.
+
+        :param algorithm: Algorithm name supported by :class:`hashlib`
+        :type algorithm: str, optional
+        :return: Checksum of file contents
+        :rtype: string object of double length, containing only hexadecimal digits
+        """
         hasher = hashlib.new(algorithm)
         hasher.update(self.content)
         return hasher.hexdigest()
@@ -229,20 +344,48 @@ class TimestreamFile(object):
     def shasum(self):
         return self.checksum('sha512')
 
+    def copy(self):
+        return copy.deepcopy(self)
+
 
 class TimeStream(object):
+    """ Represents a set of files organised in Timestream format.
+    """
     bundle_levels = ("root", "year", "month", "day", "hour", "none")
 
-    def __init__(self, path=None, format=None, onerror="warn",
+    def __init__(self, path=None, format=None,
                  bundle_level="none", name=None, timefilter=None,
                  add_subsecond_field=False, flat_output=False,
                  write_index=False):
-        """path is the base directory of a timestream"""
+        """
+        :param path: Base directory of a timestream
+        :type path: str
+        :param format: Format of files in timestream, usually the same as the file extension (e.g. ``.jpg``)
+        :type format: str
+        :param bundle_level: Smallest time unit to bundle files at.
+        :type bundle_level: str
+        :param name: Name of timestream
+        :type name: str
+        :param timefilter: Timefilter object for defining datetime range to iterate over
+        :type timefilter: :class:`time.Timefilter`, optional
+        :param add_subsecond_field: Enable for timestreams with sub-second records, using an additional ``_[00-99]`` at the end of filenames
+        :type add_subsecond_field: bool, optional
+        :param flat_output: Store timestream in a flat file structure, instead of Timestream directory structure
+        :type flat_output: bool, optional
+        :param write_index: Save file paths in timestream to an index.json file
+        :type write_index: bool, optional
+        """
         self._files = {}
         self._instants = {}
         self.name = name
         self.path = None
-        self.format = None
+        if format is not None:
+            format = format.lstrip(".").lower()
+            if format == "tiff":
+                format = "tif"
+            if format == "jpeg":
+                format = "jpg"
+        self.format = format
         self.sorted = True
         self.write_index = write_index
         self.add_subsecond_field = add_subsecond_field
@@ -253,10 +396,6 @@ class TimeStream(object):
         if bundle_level not in self.bundle_levels:
             raise ValueError("invalid bundle level %s",  bundle_level)
         self.bundle = bundle_level
-        if onerror == "raise" or onerror == "skip" or onerror == "warn":
-            self.onerror = onerror
-        else:
-            raise ValueError("onerror should be one of raise, skip, or warn")
         self._index_file = None
         if path is not None:
             self.open(path, format=format)
@@ -266,21 +405,28 @@ class TimeStream(object):
                 self._index_file = op.join(self.path, "index.json")
 
     def open(self, path, format=None):
+        """ Opens a stored timestream.
+
+        :param path: Path to timestream file/directory
+        :type path: str
+        :param format: Timestream format, if not unarchived or in a tar/zip archive
+        :type format: str, optional
+        """
         if self.name is None:
             self.name = op.basename(path)
             for ext in [".tar", ".zip", f".{format}"]:
                 if self.name.lower().endswith(ext):
                     self.name = self.name[:-len(ext)]
-        if format is not None:
-            format = format.lstrip(".").lower()
-            if format == "tiff":
-                format = "tif"
-            if format == "jpeg":
-                format = "jpg"
-        self.format = format
         self.path = path
 
     def index(self, progress=True):
+        """ Update timestream index if empty. Useful before searching a
+        timestream for specific timepoints. Also updates index file if one is
+        set/exists.
+
+        :param progress: Show progress of indexing with :class:`tqdm`
+        :type progress: bool, optional
+        """
         if len(self._instants) == 0 or len(self._files) == 0:
             with FileLock(self._index_file, timeout=3600):
                 pass
@@ -326,6 +472,13 @@ class TimeStream(object):
         return self._instants.keys()
 
     def getinstant(self, value):
+        """ Retrieves files from timestream at a specific time point.
+
+        :param value: Time point to retrieve
+        :type value: :class:`time.TSInstant`
+        :return: File if present
+        :rtype: :class:`.TimestreamFile`
+        """
         if isinstance(value, TimestreamFile):
             value = value.instant
         assert(isinstance(value, TSInstant))
@@ -337,17 +490,17 @@ class TimeStream(object):
         self.index(progress=False)
         return TimestreamFile(filename=filename, fetcher=self._files[filename])
 
-
-
     def _scan_dir(self, basedir):
         for root, dirs, files in os.walk(basedir):
             # ensure sorted iteration
             dirs.sort()
-            files.sort(key=lambda f: extract_datetime(f))
+            files.sort(key=lambda f: get_instant_or_fn(f))
             for file in files:
                 path = op.join(root, file)
                 if not op.exists(path):
                     continue
+                if file.startswith('.'):
+                    continue
                 if not path_is_timestream_file(path, extensions=self.format):
                     continue
                 fetcher = FileContentFetcher(path)
@@ -355,6 +508,12 @@ class TimeStream(object):
                 yield TimestreamFile(fetcher=fetcher)
 
     def from_inotify(self, basedir):
+        """ Watch a directory for files using ``inotify`` and yield new files as they are added to it.
+        Rescans every 10 minutes.
+
+        :param basedir: Directory to watch.
+        :type basedir: str
+        """
         import inotify.adapters
 
         while True:
@@ -365,22 +524,27 @@ class TimeStream(object):
             for _ in range(600): # rescan every 10m
                 for ev in inot.event_gen(timeout_s=1, yield_nones=False):
                     (_, type_names, path, filename) = ev
-                    if 'IN_ISDIR' in type_names or  not 'IN_CLOSE_WRITE' in type_names and not 'IN_MOVED_TO' in type_names:
+                    if 'IN_ISDIR' in type_names or not 'IN_CLOSE_WRITE' in type_names and not 'IN_MOVED_TO' in type_names:
                         continue
                     path = op.join(path, filename)
-                    if not op.exists(path):
-                        continue
-                    if not path_is_timestream_file(path, extensions=self.format):
+                    if filename.startswith('.') or not op.exists(path) or not path_is_timestream_file(path, extensions=self.format):
                         continue
                     fetcher = FileContentFetcher(path)
                     self._files[fetcher.filename] = fetcher
                     yield TimestreamFile(fetcher=fetcher)
             # clean up inotify watcher
             del inot
-    
+
 
 
     def from_fofn(self, pathorfile):
+        """ Yields files from a list of file names, most commonly from an index file.
+
+        :param pathorfile: Path to an index file, or open file of file names
+        :type pathorfile: str or :class:`io.IOBase`
+        :return: Iteratively returns files from file names
+        :rtype: :class:`.TimestreamFile`
+        """
         fp = pathorfile
         if not isinstance(pathorfile, io.IOBase):
             fp = open(pathorfile)
@@ -393,12 +557,19 @@ class TimeStream(object):
             yield TimestreamFile(fetcher=fetcher)
 
     def iter(self, tar_contents=True):
+        """ Yields files in timestream, sorted by time.
+
+        :param tar_contents: Load tar contents into TimestreamFile immediately, rather than through TarContentFetcher
+        :type tar_contents: bool, optional
+        :return: Timestream files, iteratively
+        :rtype: :class:`.TimestreamFile`
+        """
         def walk_archive(path):
             if zipfile.is_zipfile(str(path)):
                 with zipfile.ZipFile(str(path)) as zip:
                     # ensure sorted iteration
                     entries = zip.infolist()
-                    entries.sort(key=lambda entry: extract_datetime(entry.filename))
+                    entries.sort(key=lambda entry: get_instant_or_fn(entry.filename))
                     for entry in entries:
                         if entry.is_dir():
                             continue
@@ -411,7 +582,7 @@ class TimeStream(object):
                                              fetcher=ZipContentFetcher(path, entry.filename))
             elif tarfile.is_tarfile(path):
                 self.sorted = False
-                #warnings.warn("Extracting files from a tar file. Sorted iteration is not guaranteed")
+                # warnings.warn("Extracting files from a tar file. Sorted iteration is not guaranteed")
                 with tarfile.TarFile(path) as tar:
                     for entry in tar:
                         if not entry.isfile():
@@ -440,12 +611,12 @@ class TimeStream(object):
             if is_archive(self.path):
                 yield from walk_archive(self.path)
         except Exception as exc:
-            print(f"\n{exc.__class__.__name__}: {str(exc)} at '{path}'\n", file=stderr)
+            print(f"\n{exc.__class__.__name__}: {str(exc)} at '{self.path}'\n", file=stderr)
 
         for root, dirs, files in os.walk(self.path):
             # ensure sorted iteration
             dirs.sort()
-            files.sort(key=lambda f: extract_datetime(f))
+            files.sort(key=lambda f: get_instant_or_fn(f))
             for file in files:
                 path = op.join(root, file)
                 if file.startswith("."):
@@ -500,6 +671,11 @@ class TimeStream(object):
         return file.instant.datetime.strftime(bpath)
 
     def write(self, file):
+        """ Adds a file to this timestream.
+
+        :param file: A valid file for this timestream
+        :type file: :class:`.TimestreamFile`
+        """
         if op.exists(self._index_file):
             with FileLock(self._index_file):
                 try:
@@ -546,4 +722,6 @@ class TimeStream(object):
         return self.iter()
 
     def close(self):
+        """ Maintains interface as other objects with ``close()`` methods but does not currently need further actions.
+        """
         pass
